import abc
from abc import ABC

from core.requester import KibaResponse


class NotificationClient(ABC):
    @abc.abstractmethod
    async def post(self, messageText: str) -> KibaResponse:
        pass



import random
import string
from collections.abc import Sequence

CHARACTERS_ALPHANUMERIC = string.digits + string.ascii_letters


def generate_random_string(length: int, characters: Sequence[str] = CHARACTERS_ALPHANUMERIC) -> str:
    return ''.join([random.choice(characters) for _ in range(length)])  # noqa: S311



from urllib import parse


def encode_url(url: str) -> str:
    return parse.quote(string=url)


def encode_url_part(urlPart: str) -> str:
    return parse.quote(string=urlPart, safe='')


def decode_url(url: str) -> str:
    return parse.unquote(string=url)


def decode_url_part(urlPart: str) -> str:
    return parse.unquote(string=urlPart)



import contextlib
import contextvars
import typing
from collections.abc import AsyncIterator
from typing import TypeVar

from sqlalchemy.engine import Result
from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.sql.selectable import TypedReturnsRows

from core.exceptions import InternalServerErrorException

DatabaseConnection = AsyncConnection
ResultType = TypeVar('ResultType', bound=tuple)  # type: ignore[type-arg]


class Database:
    @staticmethod
    def create_connection_string(engine: str, username: str, password: str, host: str, port: str, name: str) -> str:
        return f'{engine}://{username}:{password}@{host}:{port}/{name}'

    @staticmethod
    def create_psql_connection_string(username: str, password: str, host: str, port: str, name: str) -> str:
        return Database.create_connection_string(engine='postgresql+asyncpg', username=username, password=password, host=host, port=port, name=name)

    @staticmethod
    def create_sqlite_connection_string(filename: str) -> str:
        return f'sqlite+aiosqlite:///{filename}'

    def __init__(self, connectionString: str) -> None:
        self.connectionString = connectionString
        self._engine: AsyncEngine | None = None
        self._connectionContext = contextvars.ContextVar[DatabaseConnection | None]('_connectionContext')

    async def connect(self) -> None:
        if not self._engine:
            self._engine = create_async_engine(self.connectionString, future=True)

    async def disconnect(self) -> None:
        if self._engine:
            await self._engine.dispose()
            self._engine = None

    @contextlib.asynccontextmanager
    async def create_transaction(self) -> AsyncIterator[DatabaseConnection]:
        if not self._engine:
            raise InternalServerErrorException(message='Engine has not been established. Please called collect() first.')
        async with self._engine.begin() as connection:
            yield connection

    def _get_connection(self) -> DatabaseConnection | None:
        try:
            connection = self._connectionContext.get()
            if connection and not connection.closed:
                return connection
        except LookupError:
            pass
        return None

    @contextlib.asynccontextmanager
    async def create_context_connection(self) -> AsyncIterator[DatabaseConnection]:
        if not self._engine:
            raise InternalServerErrorException(message='Engine has not been established. Please called collect() first.')
        if self._get_connection() is not None:
            raise InternalServerErrorException(message='Connection has already been established in this context.')
        async with self._engine.begin() as connection:
            self._connectionContext.set(connection)
            try:
                yield connection
            finally:
                self._connectionContext.set(None)

    async def execute(self, query: TypedReturnsRows[ResultType], connection: DatabaseConnection | None = None) -> Result[ResultType]:
        if not self._engine:
            raise InternalServerErrorException(message='Connection has not been established. Please called collect() first.')
        if connection:
            return typing.cast(Result[ResultType], await connection.execute(statement=query))
        newConnection = self._get_connection()
        if newConnection:
            return typing.cast(Result[ResultType], await newConnection.execute(statement=query))
        async with self._engine.connect() as temporaryConnection:
            return typing.cast(Result[ResultType], await temporaryConnection.execute(statement=query))



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from core.store.database import Database


class DatabaseConnectionMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, database: Database) -> None:
        super().__init__(app=app)
        self.database = database

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        if request.method in {'GET', 'OPTIONS', 'HEAD'}:
            response = await call_next(request)
        else:
            async with self.database.create_context_connection():
                response = await call_next(request)
        return response



import asyncio
import json
import typing
from typing import Any

from eth_abi.exceptions import DecodingError
from eth_abi.exceptions import InsufficientDataBytes
from eth_typing import ABI
from eth_typing import ABIFunction
from eth_utils.abi import get_abi_output_types
from web3 import Web3
from web3._utils import method_formatters
from web3._utils.contracts import encode_transaction_data
from web3._utils.rpc_abi import RPC
from web3.types import BlockData
from web3.types import HexBytes
from web3.types import HexStr
from web3.types import LogReceipt
from web3.types import Nonce
from web3.types import TxData
from web3.types import TxParams
from web3.types import TxReceipt
from web3.types import Wei

from core.exceptions import BadRequestException
from core.exceptions import KibaException
from core.exceptions import NotFoundException
from core.requester import Requester
from core.util import chain_util
from core.util.typing_util import Json

ListAny = list[Any]  # type: ignore[explicit-any]
DictStrAny = dict[str, Any]  # type: ignore[explicit-any]


class TransactionFailedException(KibaException):
    def __init__(self, transactionReceipt: TxReceipt) -> None:
        super().__init__(message='Transaction failed')
        self.transactionReceipt = transactionReceipt

    def to_dict(self) -> dict[str, Json]:
        output = super().to_dict()
        output['fields']['transactionReceipt'] = dict(self.transactionReceipt)  # type: ignore[index, call-overload]
        return output


class EthClientInterface:
    async def get_latest_block_number(self) -> int:
        raise NotImplementedError

    async def get_block(self, blockNumber: int, shouldHydrateTransactions: bool = False) -> BlockData:
        raise NotImplementedError

    async def get_block_uncle_count(self, blockNumber: int) -> int:
        raise NotImplementedError

    async def get_transaction(self, transactionHash: str) -> TxData:
        raise NotImplementedError

    async def get_transaction_receipt(self, transactionHash: str) -> TxReceipt:
        raise NotImplementedError

    async def get_log_entries(self, topics: list[str] | None = None, startBlockNumber: int | None = None, endBlockNumber: int | None = None, address: str | None = None) -> list[LogReceipt]:
        raise NotImplementedError

    async def call_function(self, toAddress: str, contractAbi: ABI, functionAbi: ABIFunction, fromAddress: str | None = None, arguments: DictStrAny | None = None, blockNumber: int | None = None) -> ListAny:
        raise NotImplementedError

    async def call_function_by_name(self, toAddress: str, contractAbi: ABI, functionName: str, fromAddress: str | None = None, arguments: DictStrAny | None = None, blockNumber: int | None = None) -> ListAny:
        raise NotImplementedError

    async def send_raw_transaction(self, transactionData: str) -> str:
        raise NotImplementedError

    async def send_transaction(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionAbi: ABIFunction,
        privateKey: str,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> str:
        raise NotImplementedError

    async def send_transaction_by_name(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionName: str,
        privateKey: str,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> str:
        raise NotImplementedError

    async def wait_for_transaction_receipt(self, transactionHash: str, sleepSeconds: int = 2) -> TxReceipt:
        raise NotImplementedError


class Web3EthClient(EthClientInterface):
    def __init__(self, web3Connection: Web3, isTestnet: bool = False) -> None:
        self.w3 = web3Connection
        self.isTestnet = isTestnet

    async def get_latest_block_number(self) -> int:
        return self.w3.eth.block_number

    async def get_block(self, blockNumber: int, shouldHydrateTransactions: bool = False) -> BlockData:
        return self.w3.eth.get_block(blockNumber, shouldHydrateTransactions)

    async def get_block_uncle_count(self, blockNumber: int) -> int:
        return self.w3.eth.get_uncle_count(blockNumber)

    async def get_transaction_count(self, address: str) -> int:
        return self.w3.eth.get_transaction_count(chain_util.normalize_address_checksum(value=address))

    async def get_transaction(self, transactionHash: str) -> TxData:
        return self.w3.eth.get_transaction(typing.cast(HexStr, transactionHash))

    async def get_transaction_receipt(self, transactionHash: str) -> TxReceipt:
        return self.w3.eth.get_transaction_receipt(typing.cast(HexStr, transactionHash))

    async def get_log_entries(
        self,
        topics: list[str] | None = None,
        startBlockNumber: int | None = None,
        endBlockNumber: int | None = None,
        address: str | None = None,
    ) -> list[LogReceipt]:
        contractFilter = self.w3.eth.filter(
            {
                'fromBlock': startBlockNumber,  # type: ignore[typeddict-item]
                'toBlock': endBlockNumber,  # type: ignore[typeddict-item]
                'topics': topics,  # type: ignore[typeddict-item]
                'address': address,  # type: ignore[typeddict-item]
            }
        )
        return contractFilter.get_all_entries()

    async def call_function(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionAbi: ABIFunction,
        fromAddress: str | None = None,
        arguments: DictStrAny | None = None,
        blockNumber: int | None = None,
    ) -> ListAny:
        raise NotImplementedError

    async def call_function_by_name(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionName: str,
        fromAddress: str | None = None,
        arguments: DictStrAny | None = None,
        blockNumber: int | None = None,
    ) -> ListAny:
        raise NotImplementedError

    async def send_raw_transaction(self, transactionData: str) -> str:
        raise NotImplementedError

    async def send_transaction(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionAbi: ABIFunction,
        privateKey: str,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> str:
        raise NotImplementedError

    async def send_transaction_by_name(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionName: str,
        privateKey: str,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> str:
        raise NotImplementedError

    async def wait_for_transaction_receipt(self, transactionHash: str, sleepSeconds: int = 2) -> TxReceipt:
        raise NotImplementedError


class RestEthClient(EthClientInterface):
    # NOTE(krishan711): find docs at https://eth.wiki/json-rpc/API
    def __init__(self, url: str, requester: Requester, isTestnet: bool = False) -> None:
        self.url = url
        self.requester = requester
        self.isTestnet = isTestnet
        self.w3 = Web3()

    @staticmethod
    def _hex_to_int(value: str) -> int:
        return int(value, 16)

    async def _make_request(self, method: str, params: ListAny | None = None) -> Any:  # type: ignore[explicit-any]
        response = await self.requester.post_json(url=self.url, dataDict={'jsonrpc': '2.0', 'method': method, 'params': params or [], 'id': 0}, timeout=100)
        jsonResponse = response.json()
        if jsonResponse.get('error'):
            raise BadRequestException(message=jsonResponse['error'].get('message') or jsonResponse['error'].get('details') or json.dumps(jsonResponse['error']))
        return jsonResponse

    async def get_latest_block_number(self) -> int:
        response = await self._make_request(method='eth_blockNumber')
        if response['result'] is None:
            raise NotFoundException
        return typing.cast(int, method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_blockNumber](response['result']))

    async def get_block(self, blockNumber: int, shouldHydrateTransactions: bool = False) -> BlockData:
        response = await self._make_request(method='eth_getBlockByNumber', params=[hex(blockNumber), shouldHydrateTransactions])
        if response['result'] is None:
            raise NotFoundException
        if self.isTestnet:
            # NOTE(krishan711): In testnet strip out the extra data as done by web3
            # https://web3py.readthedocs.io/en/stable/middleware.html#why-is-geth-poa-middleware-necessary
            response['result']['extraData'] = HexBytes('0').hex()
        return typing.cast(BlockData, method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_getBlockByNumber](response['result']))

    async def get_block_uncle_count(self, blockNumber: int) -> int:
        response = await self._make_request(method='eth_getUncleCountByBlockNumber', params=[hex(blockNumber)])
        if response['result'] is None:
            raise NotFoundException
        return typing.cast(int, method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_getUncleCountByBlockNumber](response['result']))

    async def get_transaction_count(self, address: str) -> int:
        response = await self._make_request(method='eth_getTransactionCount', params=[address, 'latest'])
        if response['result'] is None:
            raise NotFoundException
        return typing.cast(int, method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_getTransactionCount](response['result']))

    async def get_transaction(self, transactionHash: str) -> TxData:
        response = await self._make_request(method='eth_getTransactionByHash', params=[transactionHash])
        if response['result'] is None:
            raise NotFoundException
        return typing.cast(TxData, method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_getTransactionByHash](response['result']))

    async def get_transaction_receipt(self, transactionHash: str) -> TxReceipt:
        response = await self._make_request(method='eth_getTransactionReceipt', params=[transactionHash])
        if response['result'] is None:
            raise NotFoundException
        return typing.cast(TxReceipt, method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_getTransactionReceipt](response['result']))

    async def get_log_entries(
        self,
        topics: list[str] | None = None,
        startBlockNumber: int | None = None,
        endBlockNumber: int | None = None,
        address: str | None = None,
    ) -> list[LogReceipt]:
        params: DictStrAny = {
            'fromBlock': 'earliest',
        }
        if topics:
            params['topics'] = topics
        if startBlockNumber:
            params['fromBlock'] = hex(startBlockNumber)
        if endBlockNumber:
            params['toBlock'] = hex(endBlockNumber)
        if address:
            params['address'] = address
        response = await self._make_request(method='eth_getLogs', params=[params])
        return typing.cast(list[LogReceipt], method_formatters.PYTHONIC_RESULT_FORMATTERS[RPC.eth_getLogs](response['result']))

    async def call_function(
        self,
        toAddress: str,
        contractAbi: ABI,  # noqa: ARG002
        functionAbi: ABIFunction,
        fromAddress: str | None = None,
        arguments: DictStrAny | None = None,
        blockNumber: int | None = None,
    ) -> ListAny:
        data = encode_transaction_data(w3=self.w3, abi_element_identifier=functionAbi['name'], contract_abi=[functionAbi], abi_callable=functionAbi, kwargs=(arguments or {}), args=[])
        params = {
            'from': fromAddress or '0x0000000000000000000000000000000000000000',
            'to': toAddress,
            'data': data,
        }
        response = await self._make_request(method='eth_call', params=[params, blockNumber or 'latest'])
        outputTypes = get_abi_output_types(abi_element=functionAbi)
        try:
            outputData = self.w3.codec.decode(types=outputTypes, data=HexBytes(response['result']))
        except InsufficientDataBytes as exception:
            if response['result'] == '0x':
                raise BadRequestException(message=f'Empty response: {exception!s}. Maybe the method does not exist on this contract.')
            raise
        except DecodingError as exception:
            raise BadRequestException(message=str(exception))
        return list(outputData)

    def _find_abi_by_name_args(self, contractAbi: ABI, functionName: str, arguments: DictStrAny | None = None) -> ABIFunction:
        functionAbis = typing.cast(list[ABIFunction], [abi for abi in contractAbi if abi.get('name') == functionName])
        if len(functionAbis) == 0:
            raise BadRequestException(message='Function not found in ABI')
        functionAbi: ABIFunction | None = None
        if len(functionAbis) == 1:
            functionAbi = functionAbis[0]
        else:
            argumentCount = len(arguments or {})
            functionAbi = next((abi for abi in functionAbis if len(abi['inputs']) == argumentCount), None)
        if not functionAbi:
            raise BadRequestException(message='Function not found in ABI')
        return functionAbi

    async def call_function_by_name(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionName: str,
        fromAddress: str | None = None,
        arguments: DictStrAny | None = None,
        blockNumber: int | None = None,
    ) -> ListAny:
        functionAbi = self._find_abi_by_name_args(contractAbi=contractAbi, functionName=functionName, arguments=arguments)
        return await self.call_function(
            toAddress=toAddress,
            contractAbi=contractAbi,
            functionAbi=functionAbi,
            fromAddress=fromAddress,
            arguments=arguments,
            blockNumber=blockNumber,
        )

    def _get_base_transaction_params(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionAbi: ABIFunction,
        fromAddress: str,
        arguments: DictStrAny | None = None,
    ) -> TxParams:
        params: TxParams = {
            'to': chain_util.normalize_address(value=toAddress),
            'from': chain_util.normalize_address(value=fromAddress),
            'data': encode_transaction_data(w3=self.w3, abi_element_identifier=functionAbi['name'], contract_abi=contractAbi, abi_callable=functionAbi, kwargs=(arguments or {}), args=[]),
        }
        return params

    async def _get_transaction_params(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionAbi: ABIFunction,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> TxParams:
        params = self._get_base_transaction_params(toAddress=toAddress, contractAbi=contractAbi, functionAbi=functionAbi, fromAddress=fromAddress, arguments=arguments)
        if gas is None:
            response = await self._make_request(method='eth_estimateGas', params=[params])
            gas = int(response['result'], 16)
        params['gas'] = gas
        if maxPriorityFeePerGas is None:
            response = await self._make_request(method='eth_maxPriorityFeePerGas')
            maxPriorityFeePerGas = int(response['result'], 16)
        params['maxPriorityFeePerGas'] = typing.cast(Wei, maxPriorityFeePerGas)
        if maxFeePerGas is None:
            response = await self._make_request(method='eth_getBlockByNumber', params=['pending', False])
            baseFeePerGas = int(response['result']['baseFeePerGas'], 16)
            maxFeePerGas = baseFeePerGas + maxPriorityFeePerGas
        params['maxFeePerGas'] = typing.cast(Wei, maxFeePerGas)
        if nonce is None:
            nonce = await self.get_transaction_count(address=fromAddress)
        params['nonce'] = typing.cast(Nonce, nonce)
        if chainId is not None:
            params['chainId'] = chainId
        return params

    async def send_raw_transaction(self, transactionData: str) -> str:
        if not transactionData.startswith('0x'):
            transactionData = '0x' + transactionData
        response = await self._make_request(method='eth_sendRawTransaction', params=[transactionData])
        return typing.cast(str, response['result'])

    async def send_transaction(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionAbi: ABIFunction,
        privateKey: str,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> str:
        params = await self._get_transaction_params(
            toAddress=toAddress,
            contractAbi=contractAbi,
            functionAbi=functionAbi,
            fromAddress=fromAddress,
            nonce=nonce,
            gas=gas,
            maxFeePerGas=maxFeePerGas,
            maxPriorityFeePerGas=maxPriorityFeePerGas,
            arguments=arguments,
            chainId=chainId,
        )
        signedParams = self.w3.eth.account.sign_transaction(transaction_dict=params, private_key=privateKey)
        output = await self.send_raw_transaction(transactionData=signedParams.raw_transaction.hex())
        return output

    async def send_transaction_by_name(
        self,
        toAddress: str,
        contractAbi: ABI,
        functionName: str,
        privateKey: str,
        fromAddress: str,
        nonce: int | None = None,
        gas: int | None = None,
        maxFeePerGas: int | None = None,
        maxPriorityFeePerGas: int | None = None,
        arguments: DictStrAny | None = None,
        chainId: int | None = None,
    ) -> str:
        functionAbi = self._find_abi_by_name_args(contractAbi=contractAbi, functionName=functionName, arguments=arguments)
        return await self.send_transaction(
            toAddress=toAddress,
            contractAbi=contractAbi,
            functionAbi=functionAbi,
            privateKey=privateKey,
            fromAddress=fromAddress,
            nonce=nonce,
            gas=gas,
            maxFeePerGas=maxFeePerGas,
            maxPriorityFeePerGas=maxPriorityFeePerGas,
            arguments=arguments,
            chainId=chainId,
        )

    async def wait_for_transaction_receipt(self, transactionHash: str, sleepSeconds: int = 2, raiseOnFailure: bool = True) -> TxReceipt:
        while True:
            try:
                transactionReceipt = await self.get_transaction_receipt(transactionHash=transactionHash)
            except NotFoundException:
                pass
            else:
                break
            await asyncio.sleep(sleepSeconds)
        if raiseOnFailure and transactionReceipt['status'] == 0:
            raise TransactionFailedException(transactionReceipt=transactionReceipt)
        return transactionReceipt



import functools
import json
import typing
from typing import ParamSpec

from mypy_extensions import Arg
from pydantic import BaseModel
from pydantic import ValidationError

from core.api.api_request import KibaApiRequest
from core.api.api_respnse import KibaJSONResponse
from core.exceptions import BadRequestException
from core.exceptions import InternalServerErrorException

_P = ParamSpec('_P')

ApiRequest = typing.TypeVar('ApiRequest', bound=BaseModel)
ApiResponse = typing.TypeVar('ApiResponse', bound=BaseModel)


def json_route(
    requestType: typing.Type[ApiRequest],
    responseType: typing.Type[ApiResponse],
) -> typing.Callable[[typing.Callable[[Arg(KibaApiRequest[ApiRequest], 'request')], typing.Awaitable[ApiResponse]]], typing.Callable[_P, KibaJSONResponse]]:
    def decorator(func: typing.Callable[[Arg(KibaApiRequest[ApiRequest], 'request')], typing.Awaitable[ApiResponse]]) -> typing.Callable[_P, KibaJSONResponse]:
        @functools.wraps(func)
        async def async_wrapper(*args: typing.Any) -> KibaJSONResponse:  # type: ignore[explicit-any, misc]
            receivedRequest = args[0]
            pathParams = receivedRequest.path_params
            # TODO(krishan711): move these to get request only
            queryParams = receivedRequest.query_params
            bodyBytes = await args[0].body()
            if len(bodyBytes) == 0:
                body = {}
            else:
                try:
                    body = json.loads(bodyBytes.decode())
                except json.JSONDecodeError as exception:
                    raise BadRequestException(f'Invalid JSON body: {exception}')
            allParams = {**pathParams, **body, **queryParams}
            try:
                requestParams = requestType(**allParams)
            except ValidationError as exception:
                validationErrorMessage = ', '.join([f'{".".join([str(value) for value in error["loc"]])}: {error["msg"]}' for error in exception.errors()])
                raise BadRequestException(f'Invalid request: {validationErrorMessage}')
            kibaRequest: KibaApiRequest[ApiRequest] = KibaApiRequest(scope=receivedRequest.scope, receive=receivedRequest._receive, send=receivedRequest._send)  # noqa: SLF001
            kibaRequest.data = requestParams
            receivedResponse = await func(request=kibaRequest)
            if not isinstance(receivedResponse, responseType):
                raise InternalServerErrorException(f'Expected response to be of type {responseType}, got {type(receivedResponse)}')
            return KibaJSONResponse(content=receivedResponse.model_dump())

        # TODO(krishan711): figure out correct typing here
        return async_wrapper  # type: ignore[return-value]

    return decorator



from core.util import string_util


class TestStringUtil:

    def test_generate_random_string_default_characters(self):
        length = 10
        result = string_util.generate_random_string(length)
        assert len(result) == length
        assert all(c in string_util.CHARACTERS_ALPHANUMERIC for c in result)

    def test_generate_random_string_custom_characters(self):
        length = 5
        custom_chars = "ABC123"
        result = string_util.generate_random_string(length, custom_chars)
        assert len(result) == length
        assert all(c in custom_chars for c in result)

    def test_generate_random_string_zero_length(self):
        length = 0
        result = string_util.generate_random_string(length)
        assert result == ""

    def test_generate_random_string_single_character_set(self):
        length = 3
        single_char = "X"
        result = string_util.generate_random_string(length, single_char)
        assert result == "XXX"

    def test_generate_random_string_uniqueness(self):
        length = 20
        results = set(string_util.generate_random_string(length) for _ in range(5))
        assert len(results) > 1



from __future__ import annotations

import base64
import dataclasses
import datetime
import json
import typing

JwtType = typing.TypeVar('JwtType', bound='Jwt')


@dataclasses.dataclass
class Jwt:
    headerDict: dict[str, str] = dataclasses.field(default_factory=dict)
    payloadDict: dict[str, str | int] = dataclasses.field(default_factory=dict)
    signatureBytes: bytes = b''

    @property
    def headerBase64(self) -> str:  # noqa: N802
        return Jwt._base64encode_string(value=json.dumps(obj=self.headerDict, separators=(',', ':')))

    @property
    def payloadBase64(self) -> str:  # noqa: N802
        return Jwt._base64encode_string(value=json.dumps(obj=self.payloadDict, separators=(',', ':')))

    @property
    def signatureBase64(self) -> str:  # noqa: N802
        return Jwt._base64encode(value=self.signatureBytes).decode()

    @property
    def verificationString(self) -> str:  # noqa: N802
        return f'{self.headerBase64}.{self.payloadBase64}'

    def to_string(self) -> str:
        return f'{self.headerBase64}.{self.payloadBase64}.{self.signatureBase64}'

    @classmethod
    def from_string(cls: type[JwtType], jwtString: str) -> JwtType:
        # NOTE(krish): jwts should not have the b64 padding included (its lossless to remove).
        # python complains if its missing, but not if there's too much, hence the wierd command below
        jwtParts = jwtString.split('.')
        headerDict = json.loads(s=Jwt._base64decode_string(value=jwtParts[0]))
        payloadDict = json.loads(s=Jwt._base64decode_string(value=jwtParts[1]))
        signatureBytes = Jwt._base64decode(value=jwtParts[2].encode())
        return cls(headerDict=headerDict, payloadDict=payloadDict, signatureBytes=signatureBytes)

    @classmethod
    def from_jwt(cls: type[JwtType], jwt: Jwt) -> JwtType:
        return cls.from_string(jwtString=jwt.to_string())

    @staticmethod
    def _base64decode_string(value: str, shouldAddPadding: bool = True) -> str:
        return Jwt._base64decode(value=value.encode(), shouldAddPadding=shouldAddPadding).decode()

    @staticmethod
    def _base64decode(value: bytes, shouldAddPadding: bool = True) -> bytes:
        if shouldAddPadding:
            remainder = len(value) % 4
            if remainder > 0:
                value += b'=' * (4 - remainder)
        return base64.urlsafe_b64decode(value)

    @staticmethod
    def _base64encode_string(value: str, shouldRemovePadding: bool = True) -> str:
        return Jwt._base64encode(value=value.encode(), shouldRemovePadding=shouldRemovePadding).decode()

    @staticmethod
    def _base64encode(value: bytes, shouldRemovePadding: bool = True) -> bytes:
        value = base64.urlsafe_b64encode(value)
        return value.rstrip(b'=') if shouldRemovePadding else value


@dataclasses.dataclass
class KibaJwt(Jwt):
    class Key:
        ALGORITHM = 'alg'
        TYPE = 'typ'
        KEY_ID = 'kid'
        TOKEN_ID = 'tid'
        USER_ID = 'sub'
        USER_NAME = 'nam'
        ISSUE_DATE = 'iat'
        EXPIRY_DATE = 'exp'
        FINAL_REFRESH_DATE = 'frd'

    @property
    def algorithm(self) -> str | None:
        return self.headerDict.get(KibaJwt.Key.ALGORITHM)

    @property
    def keyId(self) -> str | None:  # noqa: N802
        return self.headerDict.get(KibaJwt.Key.KEY_ID)

    @property
    def tokenId(self) -> str | None:  # noqa: N802
        return typing.cast(str, self.payloadDict[KibaJwt.Key.TOKEN_ID]) if self.payloadDict.get(KibaJwt.Key.TOKEN_ID) else None

    @property
    def userId(self) -> str | None:  # noqa: N802
        return typing.cast(str, self.payloadDict[KibaJwt.Key.USER_ID]) if self.payloadDict.get(KibaJwt.Key.USER_ID) else None

    @property
    def expiryDate(self) -> datetime.datetime | None:  # noqa: N802
        expiryDateSeconds = typing.cast(int, self.payloadDict.get('exp') or 0)
        return datetime.datetime.fromtimestamp(expiryDateSeconds, tz=datetime.UTC) if expiryDateSeconds else None

    @expiryDate.setter
    def expiryDate(self, expiryDate: datetime.datetime) -> None:  # noqa: N802
        self.payloadDict[KibaJwt.Key.EXPIRY_DATE] = int(expiryDate.timestamp())

    @property
    def issueDate(self) -> datetime.datetime | None:  # noqa: N802
        issueDateSeconds = typing.cast(int, self.payloadDict.get(KibaJwt.Key.ISSUE_DATE) or 0)
        return datetime.datetime.fromtimestamp(issueDateSeconds, tz=datetime.UTC) if issueDateSeconds else None

    @issueDate.setter
    def issueDate(self, issueDate: datetime.datetime) -> None:  # noqa: N802
        self.payloadDict[KibaJwt.Key.ISSUE_DATE] = int(issueDate.timestamp())

    @property
    def finalRefreshDate(self) -> datetime.datetime | None:  # noqa: N802
        finalRefreshDateSeconds = typing.cast(int, self.payloadDict.get(KibaJwt.Key.FINAL_REFRESH_DATE) or 0)
        return datetime.datetime.fromtimestamp(finalRefreshDateSeconds, tz=datetime.UTC) if finalRefreshDateSeconds else None

    @finalRefreshDate.setter
    def finalRefreshDate(self, finalRefreshDate: datetime.datetime) -> None:  # noqa: N802
        self.payloadDict[KibaJwt.Key.FINAL_REFRESH_DATE] = int(finalRefreshDate.timestamp())



import dataclasses
import mimetypes
import os
import random
from collections.abc import AsyncGenerator
from collections.abc import Sequence
from contextlib import AsyncExitStack
from string import ascii_letters
from typing import TYPE_CHECKING
from typing import Any

from aiobotocore.session import get_session as get_botocore_session
from botocore.exceptions import ClientError
from httpx import Headers

from core import logging
from core.exceptions import InternalServerErrorException
from core.exceptions import NotFoundException
from core.util import file_util

if TYPE_CHECKING:
    from types_aiobotocore_s3 import S3Client
else:
    S3Client = Any


@dataclasses.dataclass
class PresignedUploadField:
    name: str
    value: str


@dataclasses.dataclass
class S3PresignedUpload:
    url: str
    fields: Sequence[PresignedUploadField]


@dataclasses.dataclass
class S3File:
    bucket: str
    path: str


@dataclasses.dataclass
class S3Grant:
    granteeType: str
    granteeDisplayName: str | None
    granteeURI: str | None
    permission: str


class S3Manager:
    def __init__(self, region: str, accessKeyId: str, accessKeySecret: str) -> None:
        self.region = region
        self._accessKeyId = accessKeyId
        self._accessKeySecret = accessKeySecret
        self._exitStack = AsyncExitStack()
        self._s3Client: S3Client | None = None

    async def connect(self) -> None:
        session = get_botocore_session()
        self._s3Client = await self._exitStack.enter_async_context(session.create_client('s3', region_name=self.region, aws_access_key_id=self._accessKeyId, aws_secret_access_key=self._accessKeySecret))

    async def disconnect(self) -> None:
        await self._exitStack.aclose()
        self._s3Client = None

    @staticmethod
    def _split_path_to_bucket_key(path: str) -> tuple[str, str]:
        path = path.removeprefix('s3://')
        paths = path.split('/')
        return paths[0], '/'.join(paths[1:]) if len(paths) > 0 else ''

    @staticmethod
    def _get_extra_args(accessControl: str | None = None, cacheControl: str | None = None, contentType: str | None = None, shouldUseAesEncryption: bool = False) -> dict[str, str]:
        extraArgs = {}
        if accessControl is not None:
            extraArgs['ACL'] = accessControl
        if contentType:
            extraArgs['ContentType'] = contentType
        if cacheControl:
            extraArgs['CacheControl'] = cacheControl
        if shouldUseAesEncryption:
            extraArgs['ServerSideEncryption'] = 'AES256'
        return extraArgs

    @staticmethod
    def _generate_random_filename() -> str:
        return f'random_file_{"".join(random.choice(ascii_letters) for _ in range(20))}'  # noqa: S311

    async def write_file(self, content: bytes, targetPath: str, accessControl: str | None = None, cacheControl: str | None = None, contentType: str | None = None) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        targetBucket, targetKey = self._split_path_to_bucket_key(path=targetPath)
        extraArgs = self._get_extra_args(accessControl=accessControl, cacheControl=cacheControl, contentType=contentType or self._get_file_mimetype(fileName=targetKey))
        await self._s3Client.put_object(Body=content, Bucket=targetBucket, Key=targetKey, **extraArgs)  # type: ignore[arg-type]

    async def upload_file(self, filePath: str, targetPath: str, accessControl: str | None = None, cacheControl: str | None = None, contentType: str | None = None) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        content = await file_util.read_file_bytes(filePath=filePath)
        await self.write_file(content=content, targetPath=targetPath, accessControl=accessControl, cacheControl=cacheControl, contentType=contentType)

    async def read_file(self, sourcePath: str) -> bytes:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        bucket, key = self._split_path_to_bucket_key(path=sourcePath)
        response = await self._s3Client.get_object(Bucket=bucket, Key=key)
        async with response['Body'] as stream:
            output = await stream.read()
        return output

    async def download_file(self, sourcePath: str, filePath: str) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        content = await self.read_file(sourcePath=sourcePath)
        await file_util.write_file_bytes(filePath=filePath, content=content)

    async def copy_file(self, source: str, target: str, accessControl: str | None = None, cacheControl: str | None = None, contentType: str | None = None) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        logging.info(f'Copying file from {source} to {target}')
        sourceBucket, sourceKey = self._split_path_to_bucket_key(path=source)
        targetBucket, targetKey = self._split_path_to_bucket_key(path=target)
        extraArgs = self._get_extra_args(accessControl=accessControl, cacheControl=cacheControl, contentType=contentType or self._get_file_mimetype(fileName=targetKey))
        await self._s3Client.copy_object(CopySource={'Bucket': sourceBucket, 'Key': sourceKey}, Bucket=targetBucket, Key=targetKey, MetadataDirective='REPLACE', **extraArgs)  # type: ignore[arg-type]

    async def delete_file(self, filePath: str) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        bucket, key = self._split_path_to_bucket_key(path=filePath)
        await self._s3Client.delete_object(Bucket=bucket, Key=key)

    async def head_file(self, filePath: str) -> Headers:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        bucket, key = self._split_path_to_bucket_key(path=filePath)
        try:
            response = await self._s3Client.head_object(Bucket=bucket, Key=key)
        except ClientError:
            raise NotFoundException
        return Headers(response['ResponseMetadata'].get('HTTPHeaders', {}))

    async def get_file_grants(self, filePath: str) -> Sequence[S3Grant]:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        bucket, key = self._split_path_to_bucket_key(path=filePath)
        try:
            response = await self._s3Client.get_object_acl(Bucket=bucket, Key=key)
        except ClientError:
            raise NotFoundException
        return [
            S3Grant(
                granteeType=grant['Grantee']['Type'],
                granteeDisplayName=grant['Grantee'].get('DisplayName'),
                granteeURI=grant['Grantee'].get('URI'),
                permission=grant['Permission'],
            )
            for grant in response['Grants']
        ]

    async def check_file_exists(self, filePath: str) -> bool:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        bucket, key = self._split_path_to_bucket_key(path=filePath)
        try:
            await self._s3Client.head_object(Bucket=bucket, Key=key)
        except ClientError:
            return False
        return True

    async def list_directory_files(self, s3Directory: str) -> Sequence[S3File]:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        return [s3File async for s3File in self.generate_directory_files(s3Directory=s3Directory)]

    async def generate_directory_files(self, s3Directory: str) -> AsyncGenerator[S3File]:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to read from s3')
        bucket, prefix = self._split_path_to_bucket_key(path=s3Directory)
        prefix = prefix if prefix.endswith('/') else f'{prefix}/'
        continuationToken: str | None = None
        while True:
            if not continuationToken:
                response = await self._s3Client.list_objects_v2(Bucket=bucket, Prefix=prefix, MaxKeys=100)
            else:
                response = await self._s3Client.list_objects_v2(Bucket=bucket, Prefix=prefix, MaxKeys=100, ContinuationToken=continuationToken)
            for item in response.get('Contents', []):
                yield S3File(path=item['Key'], bucket=bucket)
            if not response['IsTruncated']:
                return
            continuationToken = response['NextContinuationToken']

    @staticmethod
    def _get_file_mimetype(fileName: str) -> str | None:
        mimetype, _ = mimetypes.guess_type(fileName)
        return mimetype

    async def copy_directory(self, source: str, target: str, accessControl: str | None = None, cacheControl: str | None = None) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        logging.info(f'Copying directory from {source} to {target}')
        sourceBucket, sourceKey = self._split_path_to_bucket_key(path=source)
        sourceKey = sourceKey if sourceKey.endswith('/') else f'{sourceKey}/'
        target = target.rstrip('/')
        async for s3File in self.generate_directory_files(s3Directory=source):
            filePathPart = s3File.path.replace(sourceKey, '', 1)
            await self.copy_file(source=f's3://{sourceBucket}/{s3File.path}', target=f'{target}/{filePathPart}', accessControl=accessControl, cacheControl=cacheControl)

    async def upload_directory(self, sourceDirectory: str, target: str, accessControl: str | None = None, cacheControl: str | None = None) -> None:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        logging.info(f'Uploading directory from {sourceDirectory} to {target}')
        sourceDirectory = sourceDirectory if sourceDirectory.endswith('/') else f'{sourceDirectory}/'
        targetBucket, targetKey = self._split_path_to_bucket_key(path=target)
        for root, _, files in os.walk(sourceDirectory):
            for filePath in files:
                localFilePath = os.path.join(root, filePath)
                targetKeyPath = f'{targetKey}/{localFilePath.replace(sourceDirectory, "", 1)}'
                await self.upload_file(filePath=localFilePath, targetPath=f's3://{targetBucket}/{targetKeyPath}', accessControl=accessControl, cacheControl=cacheControl)

    async def generate_presigned_upload(self, target: str, accessControl: str | None = None, cacheControl: str | None = None, timeLimit: int = 60, sizeLimit: int = file_util.MEGABYTE) -> S3PresignedUpload:
        if not self._s3Client:
            raise InternalServerErrorException('You need to call .connect() before trying to write to s3')
        targetBucket, targetKey = self._split_path_to_bucket_key(path=target)
        # fields and conditions cannot be merged https://github.com/boto/boto3/issues/1103
        fields = {}
        conditions = [
            ['content-length-range', 1, sizeLimit],
            ['starts-with', '$Content-Type', ''],
        ]
        if accessControl:
            fields.update(self._get_extra_args(accessControl=accessControl))
            conditions.append(self._get_extra_args(accessControl=accessControl))
        if cacheControl:
            fields['Cache-Control'] = cacheControl
            conditions.append({'Cache-Control': cacheControl})
        response = await self._s3Client.generate_presigned_post(Bucket=targetBucket, Key=targetKey, Fields=fields, Conditions=conditions, ExpiresIn=timeLimit)
        return S3PresignedUpload(url=response['url'], fields=[PresignedUploadField(name=name, value=value) for name, value in response['fields'].items()])



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp


class ServerHeadersMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, name: str | None = None, version: str | None = None, environment: str | None = None) -> None:
        super().__init__(app=app)
        self.name = name
        self.version = version
        self.environment = environment

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        response = await call_next(request)
        if self.name:
            response.headers['X-Server-Name'] = self.name
        if self.version:
            response.headers['X-Server-Version'] = self.version
        if self.environment:
            response.headers['X-Server-Environment'] = self.environment
        return response



import time
import uuid

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from core import logging
from core.util.value_holder import RequestIdHolder


class LoggingMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, requestIdHolder: RequestIdHolder | None = None) -> None:
        super().__init__(app=app)
        self.requestIdHolder = requestIdHolder

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        requestId = str(uuid.uuid4()).replace('-', '')
        if self.requestIdHolder:
            self.requestIdHolder.set_value(value=requestId)
        startTime = time.time()
        logging.api(action=request.method, path=request.url.path, query=request.url.query)
        response = await call_next(request)
        duration = time.time() - startTime
        response.headers['X-Response-Time'] = str(duration)
        response.headers['X-Request-Id'] = requestId
        logging.api(action=request.method, path=request.url.path, query=request.url.query, response=response.status_code, duration=duration)
        if self.requestIdHolder:
            self.requestIdHolder.set_value(value=None)
        return response



from typing import TypeVar
from typing import Union

DictKeyType = TypeVar('DictKeyType')
DictValueType = TypeVar('DictValueType')
DictValueType2 = TypeVar('DictValueType2')


def merge_dicts(dict1: dict[DictKeyType, DictValueType], dict2: dict[DictKeyType, DictValueType2]) -> dict[DictKeyType, Union[DictValueType, DictValueType2]]:
    return {**dict1, **dict2}



import contextvars
from typing import Generic
from typing import TypeVar

T = TypeVar('T')


class ValueHolder(Generic[T]):
    def __init__(self, value: T) -> None:
        super().__init__()
        self._value = value

    def get_value(self) -> T:
        return self._value


class SettableValueHolder(ValueHolder[T]):
    def __init__(self, value: T) -> None:
        super().__init__(value=value)

    def set_value(self, value: T) -> None:
        self._value = value


class ContextSettableValueHolder(ValueHolder[T]):
    def __init__(self, defaultValue: T) -> None:
        super().__init__(value=defaultValue)
        self._valueContext = contextvars.ContextVar[T]('_valueContext')

    def set_value(self, value: T) -> None:
        self._valueContext.set(value)

    def get_value(self) -> T:
        try:
            return self._valueContext.get()
        except LookupError:
            pass
        return super().get_value()


class RequestIdHolder(ContextSettableValueHolder[str | None]):
    def __init__(self, defaultValue: str | None = None, separator: str = '-') -> None:
        super().__init__(defaultValue=defaultValue)
        self.separator = separator
        self._requestCounterContext = contextvars.ContextVar[int]('_requestCounterContext')

    def set_value(self, value: str | None) -> None:
        super().set_value(value=value)
        self._requestCounterContext.set(0)

    def get_request_id(self, shouldIncrementCounter: bool = True) -> str:
        requestCounterValue = self._requestCounterContext.get()
        requestId = f'{super().get_value()}{self.separator}{requestCounterValue}'
        if shouldIncrementCounter:
            self.increment_request_counter()
        return requestId

    def increment_request_counter(self) -> None:
        self._requestCounterContext.set(self._requestCounterContext.get() + 1)



import eth_utils
from web3 import Web3
from web3.types import ChecksumAddress

BURN_ADDRESS = '0x0000000000000000000000000000000000000000'


def normalize_address(value: str) -> str:
    return str(normalize_address_checksum(value=value))


def normalize_address_checksum(value: str) -> ChecksumAddress:
    addressBytes = eth_utils.to_bytes(hexstr=value)
    return Web3.to_checksum_address(addressBytes[-20:])



import functools
import typing
from typing import ParamSpec

from mypy_extensions import Arg
from pydantic import BaseModel

from core.api.api_request import KibaApiRequest
from core.exceptions import ForbiddenException
from core.http.jwt import Jwt

_P = ParamSpec('_P')


class Authorizer:
    async def validate_jwt(self, jwtString: str) -> Jwt:
        raise NotImplementedError


ApiRequest = typing.TypeVar('ApiRequest', bound=BaseModel)


def authorize_bearer_jwt(  # type: ignore[explicit-any]
    authorizer: Authorizer,
) -> typing.Callable[[typing.Callable[[Arg(KibaApiRequest[ApiRequest], 'request')], typing.Awaitable[typing.Any]]], typing.Callable[_P, typing.Any]]:
    def decorator(func: typing.Callable[[Arg(KibaApiRequest[ApiRequest], 'request')], typing.Awaitable[typing.Any]]) -> typing.Callable[_P, typing.Any]:  # type: ignore[explicit-any]
        @functools.wraps(func)
        async def async_wrapper(request: KibaApiRequest[ApiRequest]) -> typing.Any:  # type: ignore[explicit-any, misc]
            authorization = request.headers.get('Authorization')
            if not authorization:
                raise ForbiddenException(message='AUTH_NOT_PROVIDED')
            if not authorization.startswith('Bearer '):
                raise ForbiddenException(message='AUTH_INVALID')
            jwtString = authorization.replace('Bearer ', '')
            try:
                jwt = await authorizer.validate_jwt(jwtString=jwtString)
            except BaseException:  # noqa: BLE001
                raise ForbiddenException(message='AUTH_INVALID')
            request.authJwt = jwt
            return await func(request=request)

        # TODO(krishan711): figure out correct typing here
        return async_wrapper  # type: ignore[return-value]

    return decorator



import abc
from abc import ABC
from collections.abc import Sequence
from typing import Generic
from typing import TypeVar

from core.queues.model import Message

MessageType = TypeVar('MessageType', bound=Message)


class MessageQueue(Generic[MessageType], ABC):
    @abc.abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def disconnect(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def send_message(self, message: MessageType, delaySeconds: int = 0) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def send_messages(self, messages: Sequence[MessageType], delaySeconds: int = 0) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_message(self, expectedProcessingSeconds: int = 300, longPollSeconds: int = 0) -> MessageType | None:
        raise NotImplementedError

    @abc.abstractmethod
    async def get_messages(self, limit: int = 1, expectedProcessingSeconds: int = 300, longPollSeconds: int = 0) -> list[MessageType]:
        raise NotImplementedError

    @abc.abstractmethod
    async def delete_message(self, message: MessageType) -> None:
        raise NotImplementedError



class RestMethod:
    DELETE = 'DELETE'
    GET = 'GET'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    POST = 'POST'
    PUT = 'PUT'
    TRACE = 'TRACE'
    ALL = [DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT, TRACE]  # noqa: RUF012



from __future__ import annotations

import dataclasses
import datetime
from collections.abc import MutableMapping

from core.caching.cache import Cache
from core.util import date_util


class DictCache(Cache):
    @dataclasses.dataclass
    class CacheEntry:
        value: str
        expiryDate: datetime.datetime

    def __init__(self, isPrivate: bool = False) -> None:
        super().__init__(isPrivate=isPrivate)
        self._entries: MutableMapping[str, DictCache.CacheEntry] = {}

    async def set(self, key: str, value: str, expirySeconds: float) -> bool:
        self._entries[key] = DictCache.CacheEntry(
            value=value,
            expiryDate=date_util.datetime_from_now(seconds=expirySeconds),
        )
        return True

    def _internal_get(self, key: str) -> str | None:
        entry = self._entries.get(key)
        if not entry:
            return None
        if entry.expiryDate < date_util.datetime_from_now():
            del self._entries[key]
            return None
        return entry.value

    async def get(self, key: str) -> str | None:
        return self._internal_get(key=key)

    async def delete(self, key: str) -> bool:
        ret = self._internal_get(key=key)
        if ret:
            del self._entries[key]
        return ret is not None

    def can_store_complex_objects(self) -> bool:  # pylint: disable=no-self-use
        return True



import abc
import asyncio
import time
import urllib.parse as urlparse
import uuid
from abc import ABC
from typing import Generic
from typing import TypeVar

from core import logging
from core.exceptions import InternalServerErrorException
from core.exceptions import KibaException
from core.notifications.notification_client import NotificationClient
from core.queues.message_queue import MessageQueue
from core.queues.model import Message
from core.util.value_holder import RequestIdHolder


class MessageProcessor(ABC):
    @abc.abstractmethod
    async def process_message(self, message: Message) -> None:
        pass


class MessageNeedsReprocessingException(InternalServerErrorException):
    def __init__(self, maxRetryCount: int = 3, delaySeconds: int = 60, originalException: KibaException | None = None) -> None:
        super().__init__(message='MessageNeedsReprocessingException')
        self.maxRetryCount = maxRetryCount
        self.delaySeconds = delaySeconds
        self.originalException = originalException


MessageType = TypeVar('MessageType', bound=Message)


class MessageQueueProcessor(Generic[MessageType]):
    def __init__(self, queue: MessageQueue[MessageType], messageProcessor: MessageProcessor, notificationClients: list[NotificationClient], requestIdHolder: RequestIdHolder | None = None) -> None:
        self.queue = queue
        self.messageProcessor = messageProcessor
        self.notificationClients = notificationClients
        self.requestIdHolder = requestIdHolder

    async def _process_message(self, message: MessageType) -> None:
        requestId = message.requestId or str(uuid.uuid4()).replace('-', '')
        if self.requestIdHolder:
            self.requestIdHolder.set_value(value=requestId)
        query = urlparse.urlencode(message.content, doseq=True)
        logging.api(action='MESSAGE', path=message.command, query=query)
        startTime = time.time()
        statusCode = 200
        try:
            await self.messageProcessor.process_message(message=message)
            await self.queue.delete_message(message=message)
        except MessageNeedsReprocessingException as exception:
            logging.info(msg=f'Scheduling reprocessing for message:{message.command} due to: {exception.originalException!s}')
            await self.queue.send_message(message=message, delaySeconds=((message.postCount or 0) * exception.delaySeconds))
        except Exception as exception:  # noqa: BLE001
            statusCode = exception.statusCode if isinstance(exception, KibaException) else 500
            logging.error('Caught exception whilst processing message')
            logging.exception(exception)
            for client in self.notificationClients:
                await client.post(messageText=f'Error processing message: {message.command}\n```\n{requestId}\n{message.content}\n{exception}```')
            # TODO(krish): should possibly reset the visibility timeout
        duration = time.time() - startTime
        logging.api(action='MESSAGE', path=message.command, query=query, response=statusCode, duration=duration)
        if self.requestIdHolder:
            self.requestIdHolder.set_value(value=None)

    async def execute_batch(self, batchSize: int, expectedProcessingSeconds: int = 300, longPollSeconds: int = 20, shouldProcessInParallel: bool = False) -> int:
        logging.info('Retrieving messages...')
        messages = await self.queue.get_messages(expectedProcessingSeconds=expectedProcessingSeconds, longPollSeconds=longPollSeconds, limit=batchSize)
        if shouldProcessInParallel:
            await asyncio.gather(*[self._process_message(message=message) for message in messages])
        else:
            for message in messages:
                await self._process_message(message=message)
        return len(messages)

    async def execute(self, expectedProcessingSeconds: int = 300, longPollSeconds: int = 20) -> bool:
        processedMessageCount = await self.execute_batch(batchSize=1, expectedProcessingSeconds=expectedProcessingSeconds, longPollSeconds=longPollSeconds)
        return processedMessageCount > 0

    async def run_batches(self, batchSize: int, expectedProcessingSeconds: int = 300, longPollSeconds: int = 20, sleepTime: int = 30, totalMessageLimit: int | None = None) -> int:
        processedMessageCount = 0
        while totalMessageLimit is None or processedMessageCount < totalMessageLimit:
            innerProcessedMessageCount = await self.execute_batch(expectedProcessingSeconds=expectedProcessingSeconds, longPollSeconds=longPollSeconds, batchSize=batchSize)
            if innerProcessedMessageCount == 0:
                logging.info('No message received.. sleeping')
                await asyncio.sleep(sleepTime)
            processedMessageCount += innerProcessedMessageCount
        return processedMessageCount

    async def run(self, expectedProcessingSeconds: int = 300, longPollSeconds: int = 20, sleepTime: int = 30, totalMessageLimit: int | None = None) -> bool:
        processedMessageCount = await self.run_batches(batchSize=1, sleepTime=sleepTime, totalMessageLimit=totalMessageLimit, expectedProcessingSeconds=expectedProcessingSeconds, longPollSeconds=longPollSeconds)
        return processedMessageCount > 0



import datetime
from collections.abc import Iterator

from core.exceptions import KibaException

JSON_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'


class DateConversionException(KibaException):
    pass


def start_of_day(dt: datetime.datetime | None = None) -> datetime.datetime:
    dt = dt if dt is not None else datetime.datetime.now(tz=datetime.UTC)
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def end_of_day(dt: datetime.datetime | None = None) -> datetime.datetime:
    dt = dt if dt is not None else datetime.datetime.now(tz=datetime.UTC)
    return dt.replace(hour=23, minute=59, second=59, microsecond=999999)


def datetime_from_datetime(dt: datetime.datetime, days: int = 0, seconds: float = 0, milliseconds: int = 0, minutes: int = 0, hours: int = 0, weeks: int = 0) -> datetime.datetime:
    return dt + datetime.timedelta(days=days, seconds=seconds, milliseconds=milliseconds, minutes=minutes, hours=hours, weeks=weeks)


def datetime_from_now(days: int = 0, seconds: float = 0, milliseconds: int = 0, minutes: int = 0, hours: int = 0, weeks: int = 0) -> datetime.datetime:
    return datetime_from_datetime(dt=datetime.datetime.now(tz=datetime.UTC), days=days, seconds=seconds, milliseconds=milliseconds, minutes=minutes, hours=hours, weeks=weeks)


def datetime_from_string(dateString: str, dateFormat: str = JSON_DATE_FORMAT) -> datetime.datetime:
    try:
        dt = datetime.datetime.strptime(dateString, dateFormat).replace(tzinfo=datetime.UTC)
    except (TypeError, ValueError) as exception:
        raise DateConversionException(message=f'Invalid dateString passed to datetime_from_string: {dateString}') from exception
    return dt


def datetime_to_string(dt: datetime.datetime, dateFormat: str = JSON_DATE_FORMAT) -> str:
    return dt.strftime(dateFormat)


def datetime_to_utc(dt: datetime.datetime) -> datetime.datetime:
    if dt.tzinfo is None:
        return dt.replace(tzinfo=datetime.UTC)
    return dt.astimezone(datetime.UTC)


def date_hour_from_datetime(dt: datetime.datetime | None = None) -> datetime.datetime:
    dt = dt if dt is not None else datetime.datetime.now(tz=datetime.UTC)
    return dt.replace(minute=0, second=0, microsecond=0)


def generate_clock_hour_intervals(startDate: datetime.datetime, endDate: datetime.datetime) -> Iterator[tuple[datetime.datetime, datetime.datetime]]:
    # NOTE(krishan711) this has the results that look like [startDate, hourA:00]...[hourN:00, hourM:00]...[hourY:00, endDate]
    startDateNextHour = datetime_from_datetime(dt=date_hour_from_datetime(dt=datetime_from_datetime(dt=startDate, seconds=-1)), hours=1)
    if startDateNextHour > endDate:
        yield (startDate, endDate)
        return
    if startDate < startDateNextHour:
        yield (startDate, startDateNextHour)
    yield from generate_datetime_intervals(startDate=startDateNextHour, endDate=endDate, seconds=(60 * 60))


def generate_hourly_intervals(startDate: datetime.datetime, endDate: datetime.datetime) -> Iterator[tuple[datetime.datetime, datetime.datetime]]:
    # NOTE(krishan711) this has the results that look like [startDate, startDate + 1hr]...[startDate + N hrs, endDate]
    return generate_datetime_intervals(startDate=startDate, endDate=endDate, seconds=(60 * 60))


def generate_datetime_intervals(startDate: datetime.datetime, endDate: datetime.datetime, seconds: int) -> Iterator[tuple[datetime.datetime, datetime.datetime]]:
    counter = 1
    while startDate <= endDate:
        nextMaxDate = min(startDate + datetime.timedelta(seconds=counter * seconds), endDate)
        yield startDate, nextMaxDate
        if nextMaxDate == endDate:
            break
        startDate = nextMaxDate


def generate_dates_in_range(startDate: datetime.date, endDate: datetime.date, days: int = 1, shouldIncludeEndDate: bool = True) -> Iterator[datetime.date]:
    extraStep = 1 if shouldIncludeEndDate else 0
    startDate = startDate.date() if isinstance(startDate, datetime.datetime) else startDate
    endDate = endDate.date() if isinstance(endDate, datetime.datetime) else endDate
    for day in range(0, (endDate - startDate).days + extraStep, days):
        yield startDate + datetime.timedelta(days=day)


def calculate_diff_days(startDate: datetime.datetime, endDate: datetime.datetime) -> int:
    diffDays = (endDate - startDate).days
    return diffDays


def calculate_diff_years(startDate: datetime.datetime, endDate: datetime.datetime) -> float:
    years = endDate.year - startDate.year
    start_ordinal = startDate.replace(year=2000).toordinal()
    end_ordinal = endDate.replace(year=2000).toordinal()
    if end_ordinal < start_ordinal:
        years -= 1
        fraction = (end_ordinal + 365 - start_ordinal) / 365
    else:
        fraction = (end_ordinal - start_ordinal) / 365

    return round(years + fraction, 2)


def datetime_from_date(date: datetime.date) -> datetime.datetime:
    return datetime.datetime.combine(date=date, time=datetime.time.min, tzinfo=datetime.UTC)



from __future__ import annotations

import asyncio
import os

from core.caching.cache import Cache
from core.util import date_util
from core.util import file_util


class FileCache(Cache):
    def __init__(self, cacheDirectory: str, isPrivate: bool = False) -> None:
        super().__init__(isPrivate=isPrivate)
        self._cacheDirectory = cacheDirectory

    async def set(self, key: str, value: str, expirySeconds: float) -> bool:
        cacheFileDirectory = os.path.join(self._cacheDirectory, key)
        contentFilePath = os.path.join(cacheFileDirectory, 'content.txt')
        expiryFilePath = os.path.join(cacheFileDirectory, 'expiryDate.txt')
        expiryDateString = date_util.datetime_to_string(dt=date_util.datetime_from_now(seconds=expirySeconds))
        await asyncio.gather(
            *[
                file_util.write_file(filePath=contentFilePath, content=value),
                file_util.write_file(filePath=expiryFilePath, content=expiryDateString),
            ]
        )
        return True

    async def _internal_get(self, key: str) -> str | None:
        cacheFileDirectory = os.path.join(self._cacheDirectory, key)
        contentFilePath = os.path.join(cacheFileDirectory, 'content.txt')
        expiryFilePath = os.path.join(cacheFileDirectory, 'expiryDate.txt')
        contentFileExists = await file_util.file_exists(filePath=contentFilePath)
        if not contentFileExists:
            return None
        expiryFileExists = await file_util.file_exists(filePath=expiryFilePath)
        if not expiryFileExists:
            await file_util.remove_file(filePath=contentFilePath)
            return None
        expiryDateString = await file_util.read_file(filePath=expiryFilePath)
        expiryDate = date_util.datetime_from_string(dateString=expiryDateString)
        if expiryDate < date_util.datetime_from_now():
            await file_util.remove_file(filePath=contentFilePath)
            return None
        return await file_util.read_file(filePath=contentFilePath)

    async def get(self, key: str) -> str | None:
        return await self._internal_get(key=key)

    async def delete(self, key: str) -> bool:
        # NOTE(krishan711): we only delete content file for speed
        cacheFileDirectory = os.path.join(self._cacheDirectory, key)
        contentFilePath = os.path.join(cacheFileDirectory, 'content.txt')
        fileExists = await file_util.file_exists(filePath=contentFilePath)
        if fileExists:
            await file_util.remove_file(filePath=contentFilePath)
        return fileExists

    def can_store_complex_objects(self) -> bool:
        return False



import typing

from pydantic import BaseModel
from starlette.requests import Request

from core.http.basic_authentication import BasicAuthentication
from core.http.jwt import Jwt

ApiRequestDataType = typing.TypeVar('ApiRequestDataType', bound=BaseModel)


class KibaApiRequest(Request, typing.Generic[ApiRequestDataType]):
    data: ApiRequestDataType
    authJwt: Jwt | None = None
    authBasic: BasicAuthentication | None = None



import json
import os
import typing
import urllib.parse as urlparse
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import Sequence
from io import IOBase
from typing import IO
from typing import Any
from typing import Union

import httpx

from core import logging
from core.exceptions import HTTP_EXCEPTIONS_MAP
from core.exceptions import KibaException
from core.util import dict_util
from core.util import file_util
from core.util.typing_util import Json

KibaResponse = httpx.Response

FileContent = Union[IO[bytes], bytes]
File = Union[FileContent, tuple[str | None, FileContent]]
HttpxFileTypes = Union[
    # file (or bytes)
    FileContent,
    # (filename, file (or bytes))
    tuple[str | None, FileContent],
    # (filename, file (or bytes), content_type)
    tuple[str | None, FileContent, str | None],
    # (filename, file (or bytes), content_type, headers)
    tuple[str | None, FileContent, str | None, Mapping[str, str]],
]
HttpxRequestFiles = Union[Mapping[str, HttpxFileTypes], Sequence[tuple[str, HttpxFileTypes]]]


class ResponseException(KibaException):
    def __init__(self, message: str | None = None, statusCode: int | None = None, headers: MutableMapping[str, str] | None = None) -> None:
        super().__init__(message=message, statusCode=statusCode, exceptionType=None)
        self.headers = headers


class Requester:
    def __init__(self, headers: Mapping[str, str] | None = None, shouldFollowRedirects: bool = True) -> None:
        self.headers = headers or {}
        self.client = httpx.AsyncClient(follow_redirects=shouldFollowRedirects)

    async def get(self, url: str, dataDict: Json | None = None, data: bytes | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        return await self.make_request(method='GET', url=url, dataDict=dataDict, data=data, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def post(self, url: str, dataDict: Json | None = None, data: bytes | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        return await self.make_request(method='POST', url=url, dataDict=dataDict, data=data, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def post_json(self, url: str, dataDict: Json | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        headers = headers or httpx.Headers()
        headers.update({'Content-Type': 'application/json'})
        return await self.make_request(method='POST', url=url, dataDict=dataDict, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def post_form(
        self,
        url: str,
        formDataDict: Mapping[str, Union[str, FileContent]] | None = None,
        formFiles: Sequence[tuple[str, HttpxFileTypes]] | None = None,
        timeout: int | None = 10,
        headers: MutableMapping[str, str] | None = None,
        outputFilePath: str | None = None,
    ) -> KibaResponse:
        headers = headers or httpx.Headers()
        # headers.update({'Content-Type': 'application/x-www-form-urlencoded'})
        return await self.make_request(method='POST', url=url, formDataDict=formDataDict, formFiles=formFiles, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def patch(self, url: str, dataDict: Json | None = None, data: bytes | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        return await self.make_request(method='PATCH', url=url, dataDict=dataDict, data=data, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def patch_json(self, url: str, dataDict: Json | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        headers = headers or httpx.Headers()
        headers.update({'Content-Type': 'application/json'})
        return await self.make_request(method='PATCH', url=url, dataDict=dataDict, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def put(self, url: str, dataDict: Json | None = None, data: bytes | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        return await self.make_request(method='PUT', url=url, dataDict=dataDict, data=data, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def put_json(self, url: str, dataDict: Json | None = None, timeout: int | None = 10, headers: MutableMapping[str, str] | None = None, outputFilePath: str | None = None) -> KibaResponse:
        headers = headers or httpx.Headers()
        headers.update({'Content-Type': 'application/json'})
        return await self.make_request(method='PUT', url=url, dataDict=dataDict, timeout=timeout, headers=headers, outputFilePath=outputFilePath)

    async def make_request(
        self,
        method: str,
        url: str,
        dataDict: Json | None = None,
        data: bytes | None = None,
        formDataDict: Mapping[str, Union[str, FileContent]] | None = None,
        formFiles: Sequence[tuple[str, HttpxFileTypes]] | None = None,
        timeout: int | None = 10,
        headers: MutableMapping[str, str] | None = None,
        outputFilePath: str | None = None,
    ) -> KibaResponse:
        requestHeaders = httpx.Headers({**self.headers, **(headers or {})})
        # TODO(krishan711): rename parameter to content when ready
        contentDict = dataDict
        content = data
        if contentDict is not None:
            if content is not None:
                logging.error('Error: contentDict and content should never both be provided to make_request. content will be overwritten by contentDict.')
            if method == 'GET':
                urlParts = urlparse.urlparse(url)
                currentQuery = urlparse.parse_qs(urlParts.query)
                queryString = urlparse.urlencode(query=dict_util.merge_dicts(currentQuery, typing.cast(dict[str, str], contentDict)), doseq=True)  # type: ignore[arg-type]
                url = urlparse.urlunsplit(components=(urlParts.scheme, urlParts.netloc, urlParts.path, queryString, urlParts.fragment))
            if method in {'POST', 'PUT', 'PATCH'}:
                # TODO(krishan711): this should only happen if json is in the content headers
                # if requestHeaders.get('content-type') and requestHeaders.get('content-type').lower() == 'application/json':
                content = json.dumps(contentDict).encode()
        files: list[tuple[str, HttpxFileTypes]] | None = None
        innerData: dict[Any, Any] | None = None  # type: ignore[explicit-any]
        if formDataDict:
            if method == 'POST':
                formDataDictCleaned: dict[str, str] = {}
                files = []
                for name, value in formDataDict.items():
                    if isinstance(value, IO | bytes | IOBase):
                        files.append((name, value))
                    else:
                        formDataDictCleaned[name] = value
                innerData = formDataDictCleaned
            else:
                logging.error('Error: formDataDict should only be passed into POST requests.')
        if formFiles:
            if method == 'POST':
                files = files or []
                files += formFiles
            else:
                logging.error('Error: formFiles should only be passed into POST requests.')
        request = self.client.build_request(method=method, url=url, content=content, data=innerData, files=files, timeout=timeout, headers=requestHeaders)
        httpxResponse = await self.client.send(request=request)
        if 400 <= httpxResponse.status_code < 600:  # noqa: PLR2004
            message = httpxResponse.text
            if not message and httpxResponse.status_code == 401 and httpxResponse.headers.get('www-authenticate'):  # noqa: PLR2004
                message = httpxResponse.headers['www-authenticate']
            if HTTP_EXCEPTIONS_MAP.get(httpxResponse.status_code) is not None:
                exceptionCls = HTTP_EXCEPTIONS_MAP[httpxResponse.status_code]
                exception: KibaException = exceptionCls(message=message)
            else:
                exception = ResponseException(message=message, statusCode=httpxResponse.status_code, headers=httpxResponse.headers)
            raise exception
        # TODO(krishan711): this would be more efficient if streamed
        if outputFilePath is not None:
            if os.path.dirname(outputFilePath):
                os.makedirs(os.path.dirname(outputFilePath), exist_ok=True)
            await file_util.write_file_bytes(filePath=outputFilePath, content=httpxResponse.content)
        return httpxResponse

    async def close_connections(self) -> None:
        await self.client.aclose()



import datetime
import pytest

from core.exceptions import KibaException
from core.util import date_util


class TestDateUtil:

    def test_start_of_day_with_specific_datetime(self):
        test_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.start_of_day(dt=test_dt)
        assert result.year == 2025
        assert result.month == 2
        assert result.day == 26
        assert result.hour == 0
        assert result.minute == 0
        assert result.second == 0
        assert result.microsecond == 0
        assert result.tzinfo == datetime.UTC

    def test_start_of_day_with_none(self):
        result = date_util.start_of_day()
        now = datetime.datetime.now(tz=datetime.UTC)
        assert result.year == now.year
        assert result.month == now.month
        assert result.day == now.day
        assert result.hour == 0
        assert result.minute == 0
        assert result.second == 0
        assert result.microsecond == 0
        assert result.tzinfo == datetime.UTC

    def test_end_of_day_with_specific_datetime(self):
        test_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.end_of_day(dt=test_dt)
        assert result.year == 2025
        assert result.month == 2
        assert result.day == 26
        assert result.hour == 23
        assert result.minute == 59
        assert result.second == 59
        assert result.microsecond == 999999
        assert result.tzinfo == datetime.UTC

    def test_end_of_day_with_none(self):
        result = date_util.end_of_day()
        now = datetime.datetime.now(tz=datetime.UTC)
        assert result.year == now.year
        assert result.month == now.month
        assert result.day == now.day
        assert result.hour == 23
        assert result.minute == 59
        assert result.second == 59
        assert result.microsecond == 999999
        assert result.tzinfo == datetime.UTC

    def test_datetime_from_datetime_with_days(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(dt=base_dt, days=1)
        assert result == base_dt + datetime.timedelta(days=1)

    def test_datetime_from_datetime_with_seconds(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(dt=base_dt, seconds=30)
        assert result == base_dt + datetime.timedelta(seconds=30)

    def test_datetime_from_datetime_with_milliseconds(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(dt=base_dt, milliseconds=500)
        assert result == base_dt + datetime.timedelta(milliseconds=500)

    def test_datetime_from_datetime_with_minutes(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(dt=base_dt, minutes=15)
        assert result == base_dt + datetime.timedelta(minutes=15)

    def test_datetime_from_datetime_with_hours(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(dt=base_dt, hours=2)
        assert result == base_dt + datetime.timedelta(hours=2)

    def test_datetime_from_datetime_with_weeks(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(dt=base_dt, weeks=1)
        assert result == base_dt + datetime.timedelta(weeks=1)

    def test_datetime_from_datetime_with_multiple_parameters(self):
        base_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_from_datetime(
            dt=base_dt,
            days=1,
            seconds=30,
            milliseconds=500,
            minutes=15,
            hours=2,
            weeks=1
        )
        expected = base_dt + datetime.timedelta(
            days=1,
            seconds=30,
            milliseconds=500,
            minutes=15,
            hours=2,
            weeks=1
        )
        assert result == expected

    def test_datetime_from_now_with_delta(self):
        before = datetime.datetime.now(tz=datetime.UTC)
        result = date_util.datetime_from_now(days=1, hours=2, minutes=30)
        after = datetime.datetime.now(tz=datetime.UTC)

        expected_delta = datetime.timedelta(days=1, hours=2, minutes=30)
        assert before + expected_delta <= result <= after + expected_delta

    def test_datetime_from_string_with_default_format(self):
        test_str = "2025-02-26T14:30:45.123456"
        result = date_util.datetime_from_string(dateString=test_str)
        assert result == datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)

    def test_datetime_from_string_with_custom_format(self):
        test_str = "2025-02-26 14:30:45"
        result = date_util.datetime_from_string(dateString=test_str, dateFormat="%Y-%m-%d %H:%M:%S")
        assert result == datetime.datetime(2025, 2, 26, 14, 30, 45, tzinfo=datetime.UTC)

    def test_datetime_from_string_with_invalid_format(self):
        with pytest.raises(KibaException) as exc_info:
            date_util.datetime_from_string(dateString="invalid date")
        assert "Invalid dateString" in str(exc_info.value)

    def test_datetime_to_string_with_default_format(self):
        test_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_to_string(dt=test_dt)
        assert result == "2025-02-26T14:30:45.123456"

    def test_datetime_to_string_with_custom_format(self):
        test_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.datetime_to_string(dt=test_dt, dateFormat="%Y-%m-%d %H:%M:%S")
        assert result == "2025-02-26 14:30:45"

    def test_date_hour_from_datetime_with_specific_datetime(self):
        test_dt = datetime.datetime(2025, 2, 26, 14, 30, 45, 123456, tzinfo=datetime.UTC)
        result = date_util.date_hour_from_datetime(dt=test_dt)
        assert result.year == 2025
        assert result.month == 2
        assert result.day == 26
        assert result.hour == 14
        assert result.minute == 0
        assert result.second == 0
        assert result.microsecond == 0
        assert result.tzinfo == datetime.UTC

    def test_date_hour_from_datetime_with_none(self):
        result = date_util.date_hour_from_datetime()
        now = datetime.datetime.now(tz=datetime.UTC)
        assert result.year == now.year
        assert result.month == now.month
        assert result.day == now.day
        assert result.hour == now.hour
        assert result.minute == 0
        assert result.second == 0
        assert result.microsecond == 0
        assert result.tzinfo == datetime.UTC

    def test_generate_clock_hour_intervals_with_multiple_hours(self):
        start = datetime.datetime(2025, 2, 26, 14, 30, 0, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, 16, 45, 0, tzinfo=datetime.UTC)
        intervals = list(date_util.generate_clock_hour_intervals(startDate=start, endDate=end))
        assert len(intervals) == 3
        assert intervals[0] == (start, datetime.datetime(2025, 2, 26, 15, 0, 0, tzinfo=datetime.UTC))
        assert intervals[1] == (
            datetime.datetime(2025, 2, 26, 15, 0, 0, tzinfo=datetime.UTC),
            datetime.datetime(2025, 2, 26, 16, 0, 0, tzinfo=datetime.UTC)
        )
        assert intervals[2] == (
            datetime.datetime(2025, 2, 26, 16, 0, 0, tzinfo=datetime.UTC),
            end
        )

    def test_generate_clock_hour_intervals_with_single_hour(self):
        start = datetime.datetime(2025, 2, 26, 14, 30, 0, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, 14, 45, 0, tzinfo=datetime.UTC)
        intervals = list(date_util.generate_clock_hour_intervals(startDate=start, endDate=end))
        assert len(intervals) == 1
        assert intervals[0] == (start, end)

    def test_generate_hourly_intervals_with_multiple_hours(self):
        start = datetime.datetime(2025, 2, 26, 14, 30, 0, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, 16, 45, 0, tzinfo=datetime.UTC)
        intervals = list(date_util.generate_hourly_intervals(startDate=start, endDate=end))
        assert len(intervals) == 3
        assert intervals[0] == (start, start + datetime.timedelta(hours=1))
        assert intervals[1] == (
            start + datetime.timedelta(hours=1),
            start + datetime.timedelta(hours=2)
        )
        assert intervals[2] == (
            start + datetime.timedelta(hours=2),
            end
        )

    def test_generate_hourly_intervals_with_single_hour(self):
        start = datetime.datetime(2025, 2, 26, 14, 30, 0, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, 14, 45, 0, tzinfo=datetime.UTC)
        intervals = list(date_util.generate_hourly_intervals(startDate=start, endDate=end))
        assert len(intervals) == 1
        assert intervals[0] == (start, end)

    def test_generate_datetime_intervals_with_one_minute(self):
        start = datetime.datetime(2025, 2, 26, 14, 30, 0, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, 14, 35, 0, tzinfo=datetime.UTC)
        intervals = list(date_util.generate_datetime_intervals(startDate=start, endDate=end, seconds=60))
        assert len(intervals) == 5
        for i, interval in enumerate(intervals):
            assert interval[0] == start + datetime.timedelta(minutes=i)
            assert interval[1] == min(start + datetime.timedelta(minutes=i+1), end)

    def test_generate_datetime_intervals_with_custom_seconds(self):
        start = datetime.datetime(2025, 2, 26, 14, 30, 0, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, 14, 30, 10, tzinfo=datetime.UTC)
        intervals = list(date_util.generate_datetime_intervals(startDate=start, endDate=end, seconds=2))
        assert len(intervals) == 5
        for i, interval in enumerate(intervals):
            assert interval[0] == start + datetime.timedelta(seconds=i*2)
            assert interval[1] == min(start + datetime.timedelta(seconds=(i+1)*2), end)

    def test_generate_dates_in_range_with_default_parameters(self):
        start = datetime.date(2025, 2, 26)
        end = datetime.date(2025, 3, 1)
        dates = list(date_util.generate_dates_in_range(startDate=start, endDate=end))
        assert dates == [
            datetime.date(2025, 2, 26),
            datetime.date(2025, 2, 27),
            datetime.date(2025, 2, 28),
            datetime.date(2025, 3, 1),
        ]

    def test_generate_dates_in_range_with_custom_interval(self):
        start = datetime.date(2025, 2, 26)
        end = datetime.date(2025, 3, 1)
        dates = list(date_util.generate_dates_in_range(startDate=start, endDate=end, days=2))
        assert dates == [
            datetime.date(2025, 2, 26),
            datetime.date(2025, 2, 28),
        ]

    def test_generate_dates_in_range_without_end_date(self):
        start = datetime.date(2025, 2, 26)
        end = datetime.date(2025, 3, 1)
        dates = list(date_util.generate_dates_in_range(startDate=start, endDate=end, shouldIncludeEndDate=False))
        assert dates == [
            datetime.date(2025, 2, 26),
            datetime.date(2025, 2, 27),
            datetime.date(2025, 2, 28),
        ]

    def test_calculate_diff_days_with_positive_difference(self):
        start = datetime.datetime(2025, 2, 26, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 3, 1, tzinfo=datetime.UTC)
        assert date_util.calculate_diff_days(startDate=start, endDate=end) == 3

    def test_calculate_diff_days_with_negative_difference(self):
        start = datetime.datetime(2025, 3, 1, tzinfo=datetime.UTC)
        end = datetime.datetime(2025, 2, 26, tzinfo=datetime.UTC)
        assert date_util.calculate_diff_days(startDate=start, endDate=end) == -3

    def test_calculate_diff_years_with_whole_year(self):
        start = datetime.datetime(2025, 2, 26, tzinfo=datetime.UTC)
        end = datetime.datetime(2026, 2, 26, tzinfo=datetime.UTC)
        assert date_util.calculate_diff_years(startDate=start, endDate=end) == 1.0

    def test_calculate_diff_years_with_partial_year(self):
        start = datetime.datetime(2025, 2, 26, tzinfo=datetime.UTC)
        end = datetime.datetime(2026, 8, 26, tzinfo=datetime.UTC)
        assert date_util.calculate_diff_years(startDate=start, endDate=end) == 1.5

    def test_datetime_from_date_with_standard_date(self):
        test_date = datetime.date(2025, 2, 26)
        result = date_util.datetime_from_date(date=test_date)
        assert isinstance(result, datetime.datetime)
        assert result.year == 2025
        assert result.month == 2
        assert result.day == 26
        assert result.hour == 0
        assert result.minute == 0
        assert result.second == 0
        assert result.microsecond == 0

    def test_datetime_from_date_with_leap_year(self):
        test_date = datetime.date(2024, 2, 29)
        result = date_util.datetime_from_date(date=test_date)
        assert result.year == 2024
        assert result.month == 2
        assert result.day == 29
        assert result.hour == 0
        assert result.minute == 0
        assert result.second == 0
        assert result.microsecond == 0

    def test_datetime_to_utc_with_naive_datetime(self):
        dt = datetime.datetime(2025, 2, 26, 14, 30)
        result = date_util.datetime_to_utc(dt=dt)
        assert result == datetime.datetime(2025, 2, 26, 14, 30, tzinfo=datetime.UTC)

    def test_datetime_to_utc_with_utc_datetime(self):
        dt = datetime.datetime(2025, 2, 26, 14, 30, tzinfo=datetime.UTC)
        result = date_util.datetime_to_utc(dt=dt)
        assert result == dt

    def test_datetime_to_utc_with_non_utc_timezone(self):
        tz = datetime.timezone(datetime.timedelta(hours=2))
        dt = datetime.datetime(2025, 2, 26, 14, 30, tzinfo=tz)
        result = date_util.datetime_to_utc(dt=dt)
        assert result == datetime.datetime(2025, 2, 26, 12, 30, tzinfo=datetime.UTC)

    def test_datetime_to_utc_with_negative_offset_timezone(self):
        tz = datetime.timezone(datetime.timedelta(hours=-5))
        dt = datetime.datetime(2025, 2, 26, 14, 30, tzinfo=tz)
        result = date_util.datetime_to_utc(dt=dt)
        assert result == datetime.datetime(2025, 2, 26, 19, 30, tzinfo=datetime.UTC)
